@@ -5,6 +5,8 @@
 # pylint: disable=import-outside-toplevel
 
 from os.path import exists
+import json
+import zipfile
 from metprint import LogType, Logger, FHFormatter
 from layeredimage.layergroup import LayerGroupTypes, Layer, Group
 from layeredimage.layeredimage import LayeredImage, rasterImageOA
@@ -12,7 +14,7 @@ from layeredimage.blend import BlendType
 
 def extNotRecognised(fileName):
 	""" Output the file extension not recognised error """
-	exts = ["ora", "psd", "xcf", "pdn", "tif", "tiff", "webp", "gif", "lsr"]
+	exts = ["ora", "psd", "xcf", "pdn", "tif", "tiff", "webp", "gif", "lsr", "layered"]
 	Logger(FHFormatter()).logPrint("File extension is not recognised for file: " +
 	fileName + "! Must be " + "one of \"" + ", \"".join(exts) + "\"", LogType.ERROR)
 
@@ -29,28 +31,18 @@ def openLayerImage(file):
 	Returns:
 		LayeredImage: a layered image object
 	"""
+	functionMap = {"ora": openLayer_ORA, "psd": openLayer_PSD, "xcf": openLayer_XCF,
+	"pdn": openLayer_PDN, "tif": openLayer_TIFF, "tiff": openLayer_TIFF,
+	"webp": openLayer_WEBP, "gif": openLayer_GIF, "lsr": openLayer_LSR,
+	"layered": openLayer_LAYERED}
 	if not exists(file):
 		Logger(FHFormatter()).logPrint(file + " does not exist", LogType.ERROR)
 		raise FileExistsError
-	if compareExt(file, "ora"):
-		return openLayer_ORA(file)
-	if compareExt(file, "psd"):
-		return openLayer_PSD(file)
-	if compareExt(file, "xcf"):
-		return openLayer_XCF(file)
-	if compareExt(file, "pdn"):
-		return openLayer_PDN(file)
-	if compareExt(file, "tif") or compareExt(file, "tiff"):
-		return openLayer_TIFF(file)
-	if compareExt(file, "webp"):
-		return openLayer_WEBP(file)
-	if compareExt(file, "gif"):
-		return openLayer_GIF(file)
-	if compareExt(file, "lsr"):
-		return openLayer_LSR(file)
-	Logger(FHFormatter()).logPrint("File extension is not recognised!", LogType.ERROR)
-	extNotRecognised(file)
-	raise ValueError
+	fileExt = file.split(".")[-1]
+	if fileExt not in functionMap:
+		extNotRecognised(file)
+		raise ValueError
+	return functionMap[fileExt](file)
 
 def saveLayerImage(fileName, layeredImage):
 	"""Save a layered image to a file
@@ -59,24 +51,15 @@ def saveLayerImage(fileName, layeredImage):
 		fileName (string): path/ filename
 		layeredImage (LayeredImage): the layered image to save
 	"""
-	if compareExt(fileName, "ora"):
-		return saveLayer_ORA(fileName, layeredImage)
-	if compareExt(fileName, "psd"):
-		return saveLayer_PSD(fileName, layeredImage)
-	if compareExt(fileName, "xcf"):
-		return saveLayer_XCF(fileName, layeredImage)
-	if compareExt(fileName, "pdn"):
-		return saveLayer_PDN(fileName, layeredImage)
-	if compareExt(fileName, "tif") or compareExt(fileName, "tiff"):
-		return saveLayer_TIFF(fileName, layeredImage)
-	if compareExt(fileName, "webp"):
-		return saveLayer_WEBP(fileName, layeredImage)
-	if compareExt(fileName, "gif"):
-		return saveLayer_GIF(fileName, layeredImage)
-	if compareExt(fileName, "lsr"):
-		return saveLayer_LSR(fileName, layeredImage)
-	extNotRecognised(fileName)
-	raise ValueError
+	functionMap = {"ora": saveLayer_ORA, "psd": saveLayer_PSD, "xcf": saveLayer_XCF,
+	"pdn": saveLayer_PDN, "tif": saveLayer_TIFF, "tiff": saveLayer_TIFF,
+	"webp": saveLayer_WEBP, "gif": saveLayer_GIF, "lsr": saveLayer_LSR,
+	"layered": saveLayer_LAYERED}
+	fileExt = fileName.split(".")[-1]
+	if fileExt not in functionMap:
+		extNotRecognised(fileName)
+		raise ValueError
+	return functionMap[fileExt](fileName, layeredImage)
 
 def exportFlatImage(fileName, layeredImage):
 	""" Export the layered image to a unilayer image file """
@@ -99,7 +82,12 @@ def openLayer_ORA(file):
 	"svg:": BlendType.REFLECT, "svg:overlay": BlendType.OVERLAY,
 	"svg:difference": BlendType.DIFFERENCE,	"svg:lighten": BlendType.LIGHTEN,
 	"svg:darken": BlendType.DARKEN, "svg:screen": BlendType.SCREEN,
-	"svg:hard-light": BlendType.HARDLIGHT, "svg:soft-light": BlendType.SOFTLIGHT}
+	"svg:hard-light": BlendType.HARDLIGHT, "svg:soft-light": BlendType.SOFTLIGHT,
+	"svg:hue": BlendType.HUE, "svg:saturation": BlendType.SATURATION,
+	"svg:color": BlendType.COLOUR, "svg:luminosity": BlendType.LUMINOSITY,
+	"svg:plus": BlendType.ADDITIVE, "svg:dst-in": BlendType.DESTIN,
+	"svg:dst-out": BlendType.DESTOUT, "svg:dst-atop": BlendType.DESTATOP,
+	"svg:src-atop": BlendType.SRCATOP}
 	layersAndGroups = []
 	project = Project.load(file)
 	for layerOrGroup in project.children[::-1]:
@@ -127,7 +115,12 @@ def saveLayer_ORA(fileName, layeredImage):
 	BlendType.REFLECT: "svg:",	BlendType.OVERLAY: "svg:overlay",
 	BlendType.DIFFERENCE: "svg:difference",	BlendType.LIGHTEN: "svg:lighten",
 	BlendType.DARKEN: "svg:darken", BlendType.SCREEN: "svg:screen",
-	BlendType.SOFTLIGHT: "svg:soft-light", BlendType.HARDLIGHT: "svg:hard-light"}
+	BlendType.SOFTLIGHT: "svg:soft-light", BlendType.HARDLIGHT: "svg:hard-light",
+	BlendType.HUE: "svg:hue", BlendType.SATURATION: "svg:saturation",
+	BlendType.COLOUR: "svg:color", BlendType.LUMINOSITY: "svg:luminosity",
+	BlendType.ADDITIVE: "svg:plus", BlendType.DESTIN: "svg:dst-in",
+	BlendType.DESTOUT: "svg:dst-out", BlendType.DESTATOP: "svg:dst-atop",
+	BlendType.SRCATOP: "svg:src-atop"}
 	project = Project.new(layeredImage.dimensions[0], layeredImage.dimensions[1])
 	for layerOrGroup in layeredImage.layersAndGroups:
 		if layerOrGroup.type == LayerGroupTypes.LAYER:
@@ -155,7 +148,6 @@ def save_ORA_fix(self, path_or_file, composite_image=None, use_original=False):
 	"""
 	# This is a patch, so I'm going to need to access what I please
 	# pylint: disable=protected-access
-	import zipfile
 	from defusedxml import ElementTree as ET
 	from pyora.Render import Renderer, make_thumbnail
 	from pyora import TYPE_LAYER
@@ -187,7 +179,11 @@ def openLayer_PSD(file):
 	"overlay": BlendType.OVERLAY, "difference": BlendType.DIFFERENCE,
 	"subtract": BlendType.NEGATION, "lighten": BlendType.LIGHTEN,
 	"darken": BlendType.DARKEN, "screen": BlendType.SCREEN,
-	"soft light": BlendType.SOFTLIGHT, "hard light": BlendType.HARDLIGHT}
+	"soft light": BlendType.SOFTLIGHT, "hard light": BlendType.HARDLIGHT,
+	"exclusion": BlendType.EXCLUSION, "hue": BlendType.HUE,
+	"saturation": BlendType.SATURATION, "color": BlendType.COLOUR,
+	"luminosity": BlendType.LUMINOSITY, "divide": BlendType.DIVIDE,
+	"pin light": BlendType.PINLIGHT, "vivid light": BlendType.VIVIDLIGHT}
 	layersAndGroups = []
 	project = PSDImage.load(file)
 	for layerOrGroup in project.layers[::-1]:
@@ -222,19 +218,24 @@ def saveLayer_PSD(_fileName, _layeredImage):
 #### XCF ####
 def openLayer_XCF(file):
 	""" Open an .xcf file into a layered image """
-	from gimpformats_unofficial.gimpXcfDocument import GimpDocument
+	from gimpformats.gimpXcfDocument import GimpDocument
 	blendLookup = {0: BlendType.NORMAL, 3: BlendType.MULTIPLY,
-	7: BlendType.ADDITIVE, 17: BlendType.COLOURBURN, 16: BlendType.COLOURDODGE,
-	5: BlendType.OVERLAY, 6: BlendType.DIFFERENCE,
-	8: BlendType.NEGATION, 10: BlendType.LIGHTEN,
-	9: BlendType.DARKEN, 4: BlendType.SCREEN, 52: BlendType.XOR,
-	28: BlendType.NORMAL, 30: BlendType.MULTIPLY,
-	33: BlendType.ADDITIVE, 43: BlendType.COLOURBURN, 42: BlendType.COLOURDODGE,
-	23: BlendType.OVERLAY, 32: BlendType.DIFFERENCE,
-	34: BlendType.NEGATION, 36: BlendType.LIGHTEN,
-	35: BlendType.DARKEN, 31: BlendType.SCREEN,
-	18: BlendType.HARDLIGHT, 19: BlendType.SOFTLIGHT,
-	44: BlendType.HARDLIGHT, 45: BlendType.SOFTLIGHT}
+	4: BlendType.SCREEN, 5: BlendType.OVERLAY, 6: BlendType.DIFFERENCE,
+	7: BlendType.ADDITIVE, 8: BlendType.NEGATION, 9: BlendType.DARKEN,
+	10: BlendType.LIGHTEN, 11: BlendType.HUE, 12: BlendType.SATURATION,
+	13: BlendType.COLOUR, 14: BlendType.LUMINOSITY, 15: BlendType.DIVIDE,
+	16: BlendType.COLOURDODGE, 17: BlendType.COLOURBURN,
+	18: BlendType.HARDLIGHT, 19: BlendType.SOFTLIGHT, 20: BlendType.GRAINEXTRACT,
+	21: BlendType.GRAINMERGE, 23: BlendType.OVERLAY, 24: BlendType.HUE,
+	25: BlendType.SATURATION, 26: BlendType.COLOUR, 27: BlendType.LUMINOSITY,
+	28: BlendType.NORMAL, 30: BlendType.MULTIPLY, 31: BlendType.SCREEN,
+	32: BlendType.DIFFERENCE, 33: BlendType.ADDITIVE, 34: BlendType.NEGATION,
+	35: BlendType.DARKEN, 36: BlendType.LIGHTEN, 37: BlendType.HUE,
+	38: BlendType.SATURATION, 39: BlendType.COLOUR, 40: BlendType.LUMINOSITY,
+	41: BlendType.DIVIDE, 42: BlendType.COLOURDODGE,
+	43: BlendType.COLOURBURN, 44: BlendType.HARDLIGHT, 45: BlendType.SOFTLIGHT,
+	46: BlendType.GRAINEXTRACT, 47: BlendType.GRAINMERGE,
+	48: BlendType.VIVIDLIGHT, 49: BlendType.PINLIGHT, 52: BlendType.EXCLUSION}
 	project = GimpDocument(file)
 	# Iterate the layers and create a list of layers for each group, then remove
 	# these from the project layers
@@ -249,7 +250,8 @@ def openLayer_XCF(file):
 			while layers[index].itemPath is not None:
 				layer = layers[index]
 				groupLayers[groupIndex].append(Layer(layer.name, layer.image,
-				(layer.width, layer.height), (layer.xOffset, layer.yOffset),
+				(layer.width, layer.height),
+				(layer.xOffset - layerOrGroup.xOffset, layer.yOffset - layerOrGroup.yOffset),
 				layer.opacity, layer.visible,
 				blendModeLookup(layer.blendMode, blendLookup)))
 				layers.pop(index)
@@ -278,7 +280,7 @@ def openLayer_XCF(file):
 	return LayeredImage(layersAndGroups, (project.width, project.height))
 
 
-def saveLayer_XCF(fileName, layeredImage):
+def saveLayer_XCF(_fileName, _layeredImage):
 	""" Save a layered image as .xcf """
 	Logger(FHFormatter()).logPrint("Saving XCFs is not implemented in gimpformats - " +
 	"this is a little misleading as functions are pressent, however these are not " +
@@ -427,3 +429,65 @@ def saveLayer_LSR(fileName, layeredImage):
 	lsrImage = pylsr.LSRImage(layeredImage.dimensions,
 	fileName.split(sep)[-1].replace(".lsr", ""), layers)
 	pylsr.write(fileName, lsrImage)
+
+
+## LAYERED ##
+def openLayer_LAYERED(file):
+	""" Open a .layered file into a layered image """
+	blendLookup = {
+	"NORMAL": BlendType.NORMAL, "MULTIPLY": BlendType.MULTIPLY, "ADDITIVE": BlendType.ADDITIVE,
+	"COLOURBURN": BlendType.COLOURBURN, "COLOURDODGE": BlendType.COLOURDODGE,
+	"REFLECT": BlendType.REFLECT, "GLOW": BlendType.GLOW, "OVERLAY": BlendType.OVERLAY,
+	"DIFFERENCE": BlendType.DIFFERENCE, "NEGATION": BlendType.NEGATION, "LIGHTEN": BlendType.LIGHTEN,
+	"DARKEN": BlendType.DARKEN, "SCREEN": BlendType.SCREEN, "XOR": BlendType.XOR,
+	"SOFTLIGHT": BlendType.SOFTLIGHT, "HARDLIGHT": BlendType.HARDLIGHT,
+	"GRAINEXTRACT": BlendType.GRAINEXTRACT, "GRAINMERGE": BlendType.GRAINMERGE,
+	"DIVIDE": BlendType.DIVIDE, "HUE": BlendType.HUE, "SATURATION": BlendType.SATURATION,
+	"COLOUR": BlendType.COLOUR, "LUMINOSITY": BlendType.LUMINOSITY,
+	"PINLIGHT": BlendType.PINLIGHT, "VIVIDLIGHT": BlendType.VIVIDLIGHT,
+	"EXCLUSION": BlendType.EXCLUSION, "DESTIN": BlendType.DESTIN,
+	"DESTOUT": BlendType.DESTOUT, "DESTATOP": BlendType.DESTATOP, "SRCATOP": BlendType.SRCATOP}
+	layersAndGroups = []
+	with zipfile.ZipFile(file, 'r') as layered:
+		stack = json.load(layered.open("stack.json"))
+		# Iterate through the layers and groups
+		for layerOrGroup in stack["layersAndGroups"]:
+			if layerOrGroup["type"] == "LAYER":
+				layersAndGroups.append(grabLayer_LAYERED(layered, layerOrGroup,
+				blendLookup))
+			else:
+				# If its a group grab the layers
+				layers = [grabLayer_LAYERED(layered, layer, blendLookup) for layer in layerOrGroup["layers"]]
+				layersAndGroups.append(Group(layerOrGroup["name"], layers, layerOrGroup["dimensions"],
+				layerOrGroup["offsets"], layerOrGroup["opacity"], layerOrGroup["visible"],
+				blendModeLookup(layerOrGroup["blendmode"], blendLookup)))
+		return LayeredImage(layersAndGroups, stack["dimensions"])
+
+def grabLayer_LAYERED(zipFile, layer, blendLookup):
+	""" Grab an image from .layered """
+	from PIL import Image
+	with zipFile.open("data/" + layer["name"] + ".png") as layerImage:
+		image = Image.open(layerImage).convert('RGBA')
+	return Layer(layer["name"], image, layer["dimensions"], layer["offsets"],
+	layer["opacity"], layer["visible"], blendModeLookup(layer["blendmode"], blendLookup))
+
+
+def saveLayer_LAYERED(fileName, layeredImage):
+	""" Save a layered image as .layered """
+	import io
+	with zipfile.ZipFile(fileName, 'w') as layered:
+		layered.writestr("stack.json", json.dumps(layeredImage.json()))
+		for layer in layeredImage.layers:
+			imgByteArr = io.BytesIO()
+			layer.image.save(imgByteArr, format='PNG')
+			imgByteArr.seek(0)
+			layered.writestr("data/" + layer.name + ".png", imgByteArr.read())
+		compositeImage = layeredImage.getFlattenLayers()
+		thumbnail = compositeImage.copy()
+		thumbnail.thumbnail((256, 256))
+		imageLookup = {"composite": compositeImage, "thumbnail": thumbnail}
+		for image in imageLookup:
+			imgByteArr = io.BytesIO()
+			imageLookup[image].save(imgByteArr, format='PNG')
+			imgByteArr.seek(0)
+			layered.writestr(image + ".png", imgByteArr.read())
